# After the initial assignment of I, L, V, which has a correctness of >50%
# I think it's possible to look at the NOE connections: check whether the 
# expected NOE is observed according to the assignment
# If a residue is correctly assigned, I will expect that >50% NOE are observed
# Instead, if a residue is wrong, there may be many unexpected NOE and missing
# NOE
# In fact, this works very well, those with most peaks as expected are always
# correct assignments, therefore it serves well the role as validation
# Anyway, it may also help to determine the assignments directly ...
#
from sys import *
from string import *
import math

#------------------------------------------------------------------------------
# read in chemical shifts (predicted)
# they are categorized into I, L, and V
#
def read_shift(file,c_err,h_err):

  shifts = {}
  shifts['ALA'] = {}
  shifts['ILE'] = {}
  shifts['LEU'] = {}
  shifts['VAL'] = {}

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    entry = line.split()
    if entry[1] not in shifts.keys():
      continue
    id = int(entry[0])
    res = entry[1]
    if id not in shifts[res].keys():
      shifts[res][id] = {}
    if entry[2][0] == 'C':
      shifts[res][id][entry[2]] = float(entry[3]) + c_err
    elif entry[2][0] == 'H':
      shifts[res][id][entry[2]] = float(entry[3]) + h_err

  return shifts

#------------------------------------------------------------------------------
# read the pdb file
#
def read_coor(file):

  atoms = {}
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  m = 0
  for line in lines:
    entry = line.split()
    if entry[0] == 'TER':      # in case of polymer
      m = m + 1
      continue
    if entry[0] != 'ATOM':
      continue
    if not_methyl(entry[3],entry[2]):
      continue
    if entry[2][0] in '123':
      #id = (int(entry[4])+500*m, entry[3], entry[2][1:])
      id = (int(entry[4]), entry[3], entry[2][1:])
    else:
      #id = (int(entry[5])+500*m, entry[3], entry[2][:-1])
      id = (int(entry[5]), entry[3], entry[2][:-1])
    if id not in atoms.keys():
      atoms[id] = []
    if 'A' in entry[4]:
      coor = (float(entry[6]),float(entry[7]),float(entry[8]))
    else:
      coor = (float(entry[5]),float(entry[6]),float(entry[7]))
    atoms[id].append(coor)
  return atoms

#------------------------------------------------------------------------------
#
def not_methyl(res,atom):

  flag = 1
  if res == 'VAL' and 'HG' in atom:
    flag = 0
  if res == 'ALA' and 'HB' in atom:
    flag = 0
  if res == 'ILE' and 'HD' in atom:
    flag = 0
  if res == 'LEU' and 'HD' in atom:
    flag = 0
  return flag

#------------------------------------------------------------------------------
# calculate distance
def calc_dist(atom1, atom2):

  dist = []
  for x1,y1,z1 in atom1:
    for x2,y2,z2 in atom2:
      d = math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2))
      dist.append(d)
  dist.sort()
  return dist[0]

#------------------------------------------------------------------------------
# simulate the expected NOE peaks
#
def noe_simulate(atoms,dist_l):

  noes = {}
  for id1 in atoms.keys():
    cutoff = dist_l
    aa = id1[1]
    # peak from v/l is only half in theory
    if aa in ('VAL','LEU'):
      cutoff = 0.9*cutoff
    coor = atoms[id1]
    if id1 not in noes.keys():
      noes[id1] = []
    for id2 in atoms.keys():
      if id2 == id1:
        continue
      coor2 = atoms[id2]
      dist = calc_dist(coor,coor2)
      if dist > cutoff:
        continue
      # peak from v/l is only half in theory
      aa2 = id2[1]
      if aa2 in ('VAL','LEU') and dist > 0.9*cutoff:
        continue
      if id2 not in noes.keys():
        noes[id2] = []
      if id2 not in noes[id1]:
        noes[id1].append(id2)
        noes[id2].append(id1)
  return noes

#------------------------------------------------------------------------------
# read in experimental NOE from a formatted file
#
def read_noe(file):

  noes = {}
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    if line[0] == '#':
      continue
    entry = line.split()
    aa = entry[1]
    id = int(entry[0])
    atom = entry[2]
    nid = (id,aa,atom)
    aa2 = entry[5]
    id2 = int(entry[4])
    atom2 = entry[6]
    nid2 = (id2,aa2,atom2)
    if nid not in noes.keys():
      noes[nid] = []
    if nid2 not in noes.keys():
      noes[nid2] = []
    if nid not in noes[nid2]:
      noes[nid2].append(nid)
      noes[nid].append(nid2)
  return noes

#------------------------------------------------------------------------------
#
def sorted_insert(list,item):

  l = len(list)
  if l == 0:
    list.append(item)
  else:
    for j in range(l):
      if item[-1] > list[j][-1]:
        list.insert(j,item)
        break
    if len(list) == l:
      list.append(item)

#------------------------------------------------------------------------------
# read in the available assignments
#
def read_assign(file):

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  assigned = {}
  for line in lines:
    entry = line.split()
    if '->' not in entry:
      continue
    #if len(entry) == 0 or entry[0] == '#' or entry[0] == '?':
    #  continue
    if line[0] == '?' or line[0] == 'x':
      nid1 = (int(entry[1]),entry[2],entry[3])
      nid2 = (int(entry[5]),entry[6],entry[7])
    else:
      nid1 = (int(entry[0]),entry[1],entry[2])
      nid2 = (int(entry[4]),entry[5],entry[6])
    assigned[nid1] = nid2
  return assigned

##-----------------------------------------------------------------------
#
def check_noe_net(shift1, shift2,noe_exp,noe_pre,assigned):

  rms_h = 0.13
  rms_c = 0.1  # error for measuring cg/cb shifts
  match = 0
  diff = 0

  cbcg = {}
  for nid in assigned.keys():
    sh1 = shift1[nid[1]][nid[0]][nid[2]]
    nid2 = assigned[nid]
    sh2 = shift2[nid2[1]][nid2[0]][nid2[2]]
    diff = diff + (sh1-sh2)*(sh1-sh2)/rms_h/rms_h/25
    #diff = diff + (sh1-sh2)/rms_h/5
    if 'CB' in shift1[nid[1]][nid[0]].keys():
      shc = shift1[nid[1]][nid[0]]['CB']
    elif 'CG' in shift1[nid[1]][nid[0]].keys():
      shc = shift1[nid[1]][nid[0]]['CG']
    else:
      continue
    if nid2[0] not in cbcg.keys():
      cbcg[nid2[0]] = shc
    else:
      # use this to maximize the NOE matching first
      #diff = diff + (abs(shc-cbcg[nid2[0]])/rms_c - 2)*0.1
      # use this to correct some V/L according to CB/CG chemical shifts
      diff = diff + (abs(shc-cbcg[nid2[0]])/rms_c - 2)

  for nid in noe_exp.keys():
    if nid not in assigned.keys():
      continue
    nid2 = assigned[nid]
    noe1 = noe_exp[nid]
    noe2 = noe_pre[nid2]
    for id1 in noe1:
      if id1 not in assigned.keys():
        continue
      if assigned[id1] in noe2:
        match = match + 1

  return (match,diff)

#------------------------------------------------------------------------
#
def swap(assign,id1,id2):

  tmp = assign[id1]
  assign[id1] = assign[id2]
  assign[id2] = tmp

#------------------------------------------------------------------------------
# main

shift1 = read_shift(argv[1],0,0)   # exp
shift2 = read_shift(argv[2],0,0)   # predict
atoms = read_coor(argv[3])           # structure

noe_exp = read_noe(argv[4])
noe_pre = noe_simulate(atoms,float(argv[6]))
assigned = read_assign(argv[5])

match,diff = check_noe_net(shift1,shift2,noe_exp,noe_pre,assigned)
print match, diff

changed = 1

while changed:
  changed = 0
  for nid1 in assigned.keys():
    for nid2 in assigned.keys():
      if nid1 == nid2:
        continue
      if nid1[1] != nid2[1]:
        continue
      swap(assigned,nid1,nid2)
      match2,diff2 = check_noe_net(shift1,shift2,noe_exp,noe_pre,assigned)
      if match-diff < match2-diff2:
        nid3 = assigned[nid1]
        nid4 = assigned[nid2]
        print '  %6d %5s %5s -> %6d %5s %5s' % (nid1[0],nid1[1],nid1[2],nid3[0],
nid3[1],nid3[2])
        print '  %6d %5s %5s -> %6d %5s %5s' % (nid2[0],nid2[1],nid2[2],nid4[0],
nid4[1],nid4[2])
        print match2, diff2
        match = match2
        diff = diff2
        changed = 1
        continue
      swap(assigned,nid1,nid2)

print '#############################'
for nid1 in assigned.keys():
      nid3 = assigned[nid1]
      print '  %6d %5s %5s -> %6d %5s %5s' % (nid1[0],nid1[1],nid1[2],nid3[0],nid3[1],nid3[2])
